// common
import { RandomInt } from "../../../helpers/random";
// events
import * as EV from "../../game-events";
// trade
import { StoredResource } from "../trade/resource/resource-storage";
import { TradeNode } from "../trade/trade-node";
// city
import { City } from "./city";
import { SinglePopulationClass, POP_CLASSES } from './population-class';

/* == INTERFACES == */
export interface IPopulation {
    city: City;
    classes: Array<SinglePopulationClass>;
    classStartingPops: Array<number>;
    classPops: Array<number>;
    lastConsumptionSuccess: number;
    lastSuppliedResIds: Array<string>;
    totalPops: number;
    upgradePercent: number;
    // methods
    consumeResources(): number;
    resIsFullySupplied(resId: string): boolean;
    getTotalResConsumption(resId: string): number;
    getTotalResConsumptionList(): Array<string>;
}

/* == CITY POPS MODEL == */


export class CityPopulation implements IPopulation {
    // parent city
    city: City;
    // population fields
    classes: Array<SinglePopulationClass> = POP_CLASSES;
    classPops: Array<number> = [];
    classStartingPops: Array<number> = [];
    startingClassPops: Array<number>;
    // pop growth
    totalGrowthIncrease: number = 0;
    // supply & consumtion
    lastConsumptionSuccess: number = 0;
    lastSuppliedResIds: Array<string> = [];

    constructor(city: City) {
        this.city = city;
        this.generateClassPops();
    }

    private generateClassPops() {
        for (let classNum = 0; classNum < POP_CLASSES.length; classNum++) {
            // retrieve max starting pops and pick a random amount below
            const maxPopsForClass = POP_CLASSES_MAX_START[classNum];
            let rndClassPops: number = RandomInt.below(maxPopsForClass);
            // save starting population
            this.classStartingPops.push(rndClassPops);
            // add to class list
            this.classPops.push(rndClassPops);
        }
        this.startingClassPops = this.classPops;
    }

    getPopsOf(popClassLevel: number): number {
        return this.classPops[popClassLevel];
    }

    get totalPops(): number {
        let totalPops: number = 0;
        for (const classPops of this.classPops) { totalPops += classPops; }
        return totalPops;
    }

    increaseRandom() {
        const rndClassNum = RandomInt.below(POP_CLASSES.length);
        this.classPops[rndClassNum] += 1;
        // fire pop increased event
        this.onPopIncreased(rndClassNum);
    }

    get upgradePercent(): number {
        const currentGrowth = this.city.growth;
        return currentGrowth /POPS_UPDATE_GROWTH *100;
    }

    /* POPULATION CONSUMPTION */
    consumeResources(): number {
        // clear last consumed resource array
        this.lastSuppliedResIds = [];
        const consumed = this.getConsumedResources();
        const totalResCount = consumed.length;
        let successfulResCount: number = 0;
        for (const storedRes of consumed) {
            const consumedResAmount = storedRes.amount;
            const consumedResId = storedRes.resId;
            // market has resources
            if(this.market.hasEnough(consumedResId, consumedResAmount)) {
                successfulResCount++;
                this.market.remove(consumedResId, consumedResAmount);
                // add to successfully consumed respurces array
                this.lastSuppliedResIds.push(consumedResId);
            } else {
                // increase price anyway
                this.market.increasePrice(consumedResId, 1);
            }
        }
        const successful = successfulResCount / totalResCount;
        this.onConsumptionFinished(successful);
        return successful;
    }

    resIsFullySupplied(resId: string): boolean {
        if (this.lastSuppliedResIds.includes(resId)) { return true; }
        return false;
    }

    onConsumptionFinished(successful: number) {
        this.lastConsumptionSuccess = successful;
        this.city.growth += successful;
        this.totalGrowthIncrease += successful;
    }

    onPopIncreased(classNum: number) {
        const newClassPops = this.getAmount(classNum);
        globalThis.EVENTS.fireEvent(new EV.CityPopulationIncreasedEvent(this.city, classNum, newClassPops));
    }

    /* returns the amount of resources of "resId" that the popluation class "classNum" consumes in one turn */
    getClassResConsumption(classNum: number, resId: string): number {
        const popClass = this.classes[classNum];
        const consumesPer10 = popClass.getResourceConsumption(resId);
        const consumesSingle = consumesPer10 /10;
        const consumesTotalFloat = this.getAmount(classNum) * consumesSingle;
        return consumesTotalFloat;
    }

    getTotalResConsumption(resId: string): number {
        let consumesTotal: number = 0;
        for (const popClass of this.classes) {
            const classConsumes = this.getClassResConsumption(popClass.level, resId);
            consumesTotal += classConsumes;
        }
        return consumesTotal;
    }

    getTotalResConsumptionList(): Array<string> {
        let consumedResIds: Array<string> = [];
        for (const popClass of this.classes) {
            const classConsumed = popClass.getConsumedResIds();
            for (const classResId of classConsumed) {
                // add resId when not already in array (no doubles)
                const isNewResId = !consumedResIds.includes(classResId) && classResId;
                if(isNewResId) { consumedResIds.push(classResId); }
            }
        }
        return consumedResIds;
    }

    getConsumedResources(): Array<StoredResource> {
        let consumedResources: Array<StoredResource> = [];
        for (const consumedResId of this.getTotalResConsumptionList()) {
            const consumedAmount = this.getTotalResConsumption(consumedResId);
            const consumedRes = new StoredResource(consumedResId, consumedAmount);
            consumedResources.push(consumedRes);
        }
        return consumedResources;
    }

    /* returns the amount of taxes generated per turn from class pops */
    getTotalTaxes(): number {
        let totalTaxes: number = 0;
        for (const popClass of this.classes) {
            let classPops: number = this.getPopsOf(popClass.level);
            let classTaxes = classPops * popClass.taxRate * TAXES_MODIFIER;
            totalTaxes += classTaxes;
        }
        return Math.round(totalTaxes);
    }

    /* POPULATION GETTERS & SETTERS */
    get market(): TradeNode { return this.city.tradeNode; }

    getAmount(classNum: number): number { return this.classPops[classNum]; }
    getName(classNum: number): string { return this.classes[classNum].name; }
    setClassPops(classNum: number, classPops: number) { this.classPops[classNum] = classPops; }
}

export const POP_CLASSES_MAX_START = [4, 6, 6, 4, 2, 2];

/* == CONSTANTS == */
export const RES_CONSUMPTION_TURNS = 7;
export const POPS_UPDATE_GROWTH = 15;
export const TAXES_MODIFIER = 1;
