// common
import { Loggable } from "../../../../common/log";
// world
import { TilePath } from "../../../world/tile/tile-path";
import { IConnectedLocation } from "../../../world/location";
// entities
import { LocationConnectionPath } from "./connection-path";
// trade
import { MarketTrade } from "../market-trade";
// babylon
import * as BABYLON from 'babylonjs';

/* == CONNECTIONS INTERFACES == */
export interface IConnection {
    // base properties
    source: IConnectedLocation;
    target: IConnectedLocation;
    level: number;
    worldDistance: number;
    tileDistance: number;
    // optional properties
    path: LocationConnectionPath;
    tilePath: TilePath;
    doneTrades: number;
    // averages
    averageWealth: number;
    // methods
    hasTarget(target: IConnectedLocation): boolean;
}

/* == CONNECTION MODEL == */
export class TradeConnection extends Loggable implements IConnection {
    // basic properties
    source: IConnectedLocation;
    target: IConnectedLocation;
    level: number = 2;
    // trade related properties
    doneTrades: number = 0;
    tradedCoins: number = 0;
    // tiles & path
    tileDistance: number;
    path: LocationConnectionPath;
    tilePath: TilePath = null;

    constructor(source: IConnectedLocation, target: IConnectedLocation, findPath?: boolean) {
        super("CityConnection");
        this.source = source;
        this.target = target;
        // find distances
        this.tileDistance = this._findTileDistance();
        // find path if required
        this.path = new LocationConnectionPath(this, source, target);
    }

    private _findTileDistance(): number {
        const distX = Math.abs(this.source.pos.x - this.target.pos.x);
        const distZ = Math.abs(this.source.pos.z - this.target.pos.z);
        return distX + distZ;
    }

    get worldDistance(): number {
        const sourceMesh = this.source.mesh;
        const targetMesh = this.target.mesh;
        if(! (sourceMesh && targetMesh)) {
            this.warn(`cant find distance between city meshes ${sourceMesh} and ${targetMesh}`);
            return -1;
        }
        const worldDist =  BABYLON.Vector3.Distance(sourceMesh.position, targetMesh.position);
        return Math.round(worldDist);
    }

    get locations(): IConnectedLocation[] {
        return [this.source, this.target];
    }

    getDistance(distanceType: number): number {
        if (distanceType == DISTANCE_TYPE_TILES) {
            return this.tileDistance;
        } else {
            return this.worldDistance;
        }
    }

    get mirrored(): TradeConnection {
        if(globalThis.CONN) {
            return globalThis.CONN.getMirrored(this);
        }
        return;
    }

    get averageWealth(): number {
        return (this.source.money + this.target.money) /2;
    }

    addToTradeHistory(marketTrade: MarketTrade, updateMirrored?: boolean): void {
        this.doneTrades += 1;
        this.tradedCoins += marketTrade.profit;
        if (updateMirrored && updateMirrored == true) {
            this.mirrored.doneTrades += 1;
            this.mirrored.tradedCoins += marketTrade.profit;
        }
    }

    hasTarget(target: IConnectedLocation): boolean {
        if (this.target.key === target.key) { return true; }
        return false;
    }

    /* CONSTANTS */
    static get MAX_DISTANCE() { return 12; }

}

/* == CONNECTION CONSTANTS == */
export const DISTANCE_TYPE_TILES = 0;
export const DISTANCE_TYPE_WORLD = 1;
