// common
import { RandomInt } from "../../../helpers/random";
// models
import { Size } from "../../world/pos";
// gui
import { AbstractGuiElement, IGuiElement } from "./element";
import { Switcher } from "../elements/switcher";
import * as GUI_CONST from "./constants";
import { SingleDataRow } from "../elements/data-row";
// babylon
import * as GUI from 'babylonjs-gui';
import { DawnbringerPalette } from "../../../common/colors";
import { GuiLabel } from "../elements/babylon/labels";
import { SectionHeader } from "../elements/section-header";
import { GuiTools } from "./gui-tools";

export interface IGuiContainer extends IGuiElement {
    babylon: GUI.Container;
    maxSize: Size;
    build();
    addCustomContent();
    updateCustomContent();
    addControl(control: GUI.Control, customSize?: boolean): GUI.Control;
    // children
    children: Array<IGuiElement>;
    childrenIds: Array<string>;
    add(elem: IGuiElement): void;
    clear(): void;
    // on events
    onElementAdded(elem: IGuiElement): void;
    onCleared(): void;
}

export class AbstractGuiContainer extends AbstractGuiElement implements IGuiContainer {
    babylon: GUI.Container;
    id: string;
    parent: IGuiElement;
    // layout properties
    children: Array<IGuiElement> = [];
    childrenIds: Array<string> = [];
    dir: number;
    maxSize: Size = null;
    private dividers = 0;
    // defaults
    forceEmpty: boolean = false;

    constructor(id: string) {
        super(id);
    }

    add(elem: IGuiElement) {
        this.log(`adding ${elem}`);
        if (this.forceEmpty) {
            this.error(`cant add ${elem}; is forced empty`);
        }
        this.childrenIds.push(elem.id);
        this.children.push(elem);
        this.onElementAdded(elem);
    }

    addMultiple(elements: Array<IGuiElement>) {
        for (const elem of elements) {
            this.add(elem);
        }
    }

    protected nextDividerNum(): number {
        this.dividers += 1;
        return this.dividers;
    }

    /* CONTENT */
    addCustomContent() {
        // overwrite this method to add any BABYLON Controls as children
        this.warn(`not adding any content; have you overwritten addCustomContent()?`);
    }

    updateCustomContent() {
        // overwrite this mission to update content: per example change a label's text
    }

    clear(): void {
        for (const child of this.children) {
            child.dispose();
        }
        this.childrenIds = [];
        this.children = [];
        this.onCleared();
    }

    addControl(control: GUI.Control): GUI.Control {
        return this.babylon.addControl(control);
    }

    /* ON CHANGE */
    onElementAdded(elem: IGuiElement) {
        elem.parent = this;
        this.log(`added element "${elem.id}"`);
    }

    onElementClicked(elemId: string): boolean {
        this.log(`click on element "${elemId}"`);
        return false;
    }

    onCleared() {
        this.log("cleared");
    }
}

export class FlexGuiLayout extends AbstractGuiContainer implements IGuiContainer {
    title: string = null;
    outter: GUI.ScrollViewer;

    constructor(id: string, dir: number, title?: string, maxSize?: Size, parent?: IGuiContainer) {
        super(id);
        this.dir = dir;
        // optional properties
        if (title) { this.title = title; }
        if (maxSize) { this.maxSize = maxSize; }
        // build!
        this.init();
    }

    init(): void { }

    build(isWide?: boolean): void {
        // create empty StackPanel
        this.babylon = this.buildContainer(isWide);
        // add header label with title text
        this.addDialogHeader();
        // add method to allow custom layout content by overwriting it
        this.addCustomContent();
        this.adjustOutterSize();
    }

    private adjustOutterSize() {
        this.outter.widthInPixels = this.babylon.widthInPixels + 40;
        this.outter.height = "95%";
        this.outter.horizontalBar.parent.isVisible = false;
        this.outter.verticalBar.parent["thickness"] = 0;
    }

    addControl(control: GUI.Control, customSize?: boolean): GUI.Control {
        const container = this.babylon as GUI.Container;
        if (! container) {
            this.error(`cant add "${control.name}"; container StackPanel ${this.babylon} not initialized`);
            return;
        }
        // custom size?
        if (! customSize) {
            control.height = GUI_CONST.DEFAULT_ELEMENT_HEIGHT;
        }
        return container.addControl(control);
    }

    protected buildContainer(isWide?: boolean): GUI.Container {
        // build outter
        this.outter = new GUI.ScrollViewer(this.id + "-outter");
        this.outter.scrollBackground = "#00000000";
        this.outter.thickness = 0;
        this.outter.height = "90%";
        if (isWide) { this.outter.width = GUI_CONST.DIALOG_OUTTER_WIDTH; }
        else { this.outter.widthInPixels = GUI_CONST.DEFAULT_INNER_WIDTH_PX + 40; }
        // build container
        let container: GUI.StackPanel = new GUI.StackPanel(this.id);
        if (isWide) { container.widthInPixels = GUI_CONST.DIALOG_OUTTER_WIDTH_PX; }
        else { container.width = GUI_CONST.DEFAULT_INNER_WIDTH; }
        // add container to outter & outte to gui root
        this.outter.addControl(container);
        globalThis.GUI.advTexture.addControl(this.outter);
        return container;
    }

    /* == ADD STANDARD ELEMENTS == */
    protected addDialogHeader() {
        if (! this.title) { return; }
        // create container
        const outter = new GUI.Rectangle(`${this.id}-header`);
        outter.width = GUI_CONST.DEFAULT_WIDTH;
        outter.height = GUI_CONST.DEFAULT_ELEMENT_HEIGHT;
        outter.thickness = 0;
        // background image
        const bgImage = new GUI.Image(`${this.id}-img`, GUI_CONST.BG_HEADER);
        bgImage.height = GUI_CONST.DEFAULT_ELEMENT_HEIGHT;
        // actual label
        const label = new GUI.TextBlock(this.id, this.title);
        label.height = GUI_CONST.DEFAULT_ELEMENT_HEIGHT;
        label.fontSize = GUI_CONST.TEXTSIZE_BIGGEST;
        label.fontFamily = GUI_CONST.FONT_HIGHLIGHTED;
        // color: dawnbringer golden fizz
        const hexColor = BABYLON.Color3.FromArray(DawnbringerPalette.GOLDENFIZZ).toHexString();
        label.color = hexColor;
        // add to container
        outter.addControl(bgImage);
        outter.addControl(label);
        this.addControl(outter);
        this.addNewDivider(true);
    }

    addNewLabel(labelId: string, text: string, clickable?: boolean, smallHeight?: boolean, noBackground?: boolean, setParent?: GUI.Container): GUI.TextBlock {
        let parent: GUI.Container = this.babylon as GUI.Container;
        if ( setParent ) { parent = setParent; }
        const labelElem = GuiTools.addNewGuiLabel(parent, labelId, text, clickable, smallHeight, noBackground);
        // return actual label
        return labelElem.labelTextBlock;
    }

    addNewDivider(empty?: boolean, setParent?: GUI.Container): GUI.TextBlock {
        let parent: GUI.Container = this.babylon as GUI.Container;
        if ( setParent ) { parent = setParent; }
        let dividerNum: number = this.nextDividerNum();
        let dividerText = "";
        if (! empty) { dividerText = "--------"; }
        let dividerLabel: GUI.TextBlock = this.addNewLabel(`divider-${dividerNum}`, dividerText, false, true, false, parent);
        dividerLabel.width = GUI_CONST.DEFAULT_INNER_WIDTH;
        dividerLabel.textVerticalAlignment = GUI_CONST.ALIGN_V_CENTER;
        return dividerLabel;
    }

    addNewSwitcher( elemId: string, options: Array<string>) {
        const switcher = new Switcher( elemId, options );
        this.addControl( switcher );
        this.addNewDivider(true);
    }

    addNewSectionHeader( headerName: string ) {
      const sectionHeader = new SectionHeader(headerName);
      this.addControl(sectionHeader, true);
    }

    addNewDataRow(key: string, value: string, inlineIcon?: string, clickable?: boolean, overwriteElemId?: string, overwriteParent?: GUI.Container): SingleDataRow {
        const layout = this;
        const dataRow = new SingleDataRow(key, value, inlineIcon, overwriteElemId, clickable);
        if ( clickable ) {
            dataRow.onPointerDownObservable.add( () => {
                const elemId = dataRow.name;
                if (layout) {layout.onElementClicked(elemId); }
            });
        }
        if (! overwriteParent) { this.addControl( dataRow, true ); }
        else { overwriteParent.addControl(dataRow); }
        return dataRow;
    }

    onElementClicked(elemId: string): boolean {
        if (! this) { return false; }
        if (!elemId ) { return false; }
        this.log(`non button clicked: "${elemId}"`);
        return true;
    }
}
