// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// babylon
import * as BABYLON from 'babylonjs';
import 'babylonjs-loaders';
import { Asset, IAsset } from '../../common/asset';

/* INTERFACES */

interface IAssetManager {
    assets: Array<IAsset>;
    loadAssets(): void;
}

/* ASSET MANAGER */
export class AssetsLoader extends BaseEventListener implements IAssetManager {
    private assetNames: Array<string> = [];
    public assets: Array<Asset> = [];
    private assetsToLoad: number = 0;
    private assetsLoaded: number = 0;

    constructor() {
        super("AssetManager");
        this.log("AssetsLoader created");
    }

    loadAssets() {
        const assetsCount: number = this.assets.length;
        this.log(`initializing loading of ${assetsCount} assets ..`);
        this.createEmptyAssets();
        // start model loading
        let loadingAssetCount: number = 0;
        for (const asset of this.assets) {
            const loadingStarted = this.loadSingleAsset(asset);
            if ( loadingStarted ) { loadingAssetCount += 1; }
        }
        this.log(`loading of ${loadingAssetCount} assets started`);
    }

    private createEmptyAssets(): void {
        for (const modelFileName of MODELS_TO_IMPORT) {
            const empty = new Asset(modelFileName);
            // add empty asset
            this.assets.push(empty);
            this.assetNames.push(empty.name);
            // this.startModelLoading(modelFileName);
            this.assetsToLoad += 1;
        }
    }

    loadSingleAsset(asset: Asset): boolean {
        const modelFileName = asset.fileName;
        // this.log(`start loading of ${modelFileName} ..`);
        if (! globalThis.SCENE) {
            this.error(`cant load ${modelFileName}; scene not yet initialized`);
            return false;
        }
        const rootDir = MODEL_DIRECTORY;
        const assetName = modelFileName.split(".")[0];
        BABYLON.SceneLoader.ImportMesh(assetName, rootDir, modelFileName, globalThis.SCENE,
            (newMeshes: Array<BABYLON.AbstractMesh>) => {
                // console.log(`LOADED MESH ${assetName} meshes "${newMeshes.length}`);
                globalThis.ASSETS.onSingleAssetLoaded(assetName, newMeshes);
            }),
            (error) => {
                console.error(error);
            }
        return true;
    }

    onSingleAssetLoaded(assetName: string, meshes: Array<BABYLON.AbstractMesh>) {
        this.log(`model ${assetName} loaded: ${meshes.length} meshes`);
        const loadedAsset = this.getAsset(assetName);
        const assetsMgr = this;
        meshes[0].name = "asset-" + assetName;
        if (meshes.length > 1) { meshes[1].id = assetName + "-1"; }
        loadedAsset.setMeshes(meshes);
        // renew loading counters
        // this.assetsLoaded += 1;
        this.assetsToLoad -= 1;
        if (this.allAssetsLoaded) {
            assetsMgr.onAllAssetsLoaded();
        }
    }

    onAllAssetsLoaded() {
        this.log("all assets loaded!");
        // before world placement
        // globalThis.FACTIONS.onBeforeWorldPlacement();
        // inform managers
        globalThis.WORLD.onAllAssetsLoaded();
        // fire assets loaded event
        globalThis.EVENTS.fireEvent( new EV.AssetsLoadedEvent(this.assetsLoaded) );
    }

    get allAssetsLoaded(): boolean {
        return this.assetsToLoad == 0;
    }

    getAsset(assetName: string, autoscale?: boolean): Asset {
        if (autoscale != true) {
            autoscale = false;
        }
        const index = this.assetNames.indexOf(assetName);
        const asset = this.assets[index];
        if (!asset) {
            this.error(`cant get asset ${assetName}: is unknown`); return;
        } else if (asset.errorMessage && asset.errorMessage.length > 0) {
            this.error(`cant get asset ${assetName}: loading error: ${asset.errorMessage}`);
            return;
        }
        return asset;
    }
}

/* == MODEL NAME CONSTANTS == */
export const CURSOR_MODELS = [ "cursor_highlighted.obj", "cursor_move_green.obj", "cursor_move_red.obj" ];

export const CITY_MODELS = [
    "rural_townhall.obj", "city_asian.obj", "city_tribal.obj", "city_venetian.obj", "city_hanseatic.obj" , "city_hellenic.obj",
    "city_mideast.obj", "city_slavic.obj", "city_weur.obj", "city_elven.obj", "city_colonial.obj",
    "city_dwarven.obj", "city_barbarian.obj", "city_cloister.obj", "city_roman.obj", "city_evil.obj",
    // unique citiy models
    "city_vivec.obj"
];

export const ARMY_MODELS = [
    "rider.obj", "ship_smallest.obj", "army_camp.obj", "army_fort.obj"
];

export const FEATURE_MODELS = [
    "forest.obj", "treasure.obj", "volcano.obj", "temple.obj", "tipis.obj", "shipwreck.obj", // features 1
    "village.obj", "oasis.obj", "arctic_peak.obj", "shrooms.obj", "ruins.obj", "watchtower.obj", // features 2
    "watchtower_white.obj", "farmstead.obj", "ice_floe.obj", "lighttower.obj", "pyramids.obj", "island.obj" // features 3
];

export const MODELS_TO_IMPORT = CURSOR_MODELS.concat(FEATURE_MODELS, CITY_MODELS, ARMY_MODELS);

/* == ASSET LOADER CONSTANTS == */
export const MODEL_DIRECTORY = "./res/models/";