// common
import { RandomElement } from "../../helpers/random";
// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// entities
import { City } from "../entities/city/city";
import { Character, ICharacter } from "../entities/character/character";
// manager
import { AbstractManager } from "./abstract-manager";

export class CharacterManager extends AbstractManager {
    nameList: {} = null;
    characters: Array<Character> = [];
    hasCreatedLeaders: boolean = false;

    constructor() {
        super("CharManager");
    }

    /* INITIALIZATION */
    onLanguagesLoaded(nameListObj: {}) {
        this.log("Character name list initialized");
        this.nameList = nameListObj;
        // inform faction manager
        globalThis.FACTIONS.onNameListLoaded();
        // create leaders
        if (! this.hasCreatedLeaders) { this.createCityLeaders(); }
        
    }

    get hasNameListLoaded(): boolean {
        if ( this.nameList ) {
            const languagesCount = Object.keys(this.nameList).length;
            if ( languagesCount > 1 ) { return true; }
        }
        return false;
    }

    createPlayerCharacter() {

    }

    createCityLeaders() {
        if(this.hasCreatedLeaders) { return; }
        if(! this.hasNameListLoaded) {
            this.warn(`cant create city leaders; name list not loaded yet`);
            return;
        }
        const citiesCount = globalThis.CITIES.citiesCount;
        let createdCount: number = 0;
        this.log(`creating leaders for ${citiesCount} cities`);
        for (const city of globalThis.CITIES.getAll()) {
            this.createCharacterIn(city, true);
            createdCount += 1;
        }
        this.log(`created ${createdCount} character leaders in cities`);
        this.hasCreatedLeaders = true;
    }

    /* CHARACTER GETTERS */
    getByNum(charNum: number): Character {
        const characterObj = this.characters[charNum];
        return characterObj;
    }

    getRandom(): Character {
        return RandomElement.of(this.characters) as Character;
    }

    getRandomName(language: string, isFemale?: boolean): string {
        if (! this.hasNameListLoaded) {
            this.warn(`cant generaate random name; name list not found`);
            return;
        }
        const languageNameList = this.nameList[language];
        // family name
        let rndFamilyName = "";
        if (languageNameList[2].length > 0) { rndFamilyName = RandomElement.of(languageNameList[2]); }
        // own name
        let rndCharName: string = RandomElement.of(languageNameList[0]);
        if (isFemale) { rndCharName = RandomElement.of(languageNameList[1]); }
        return rndCharName + " " + rndFamilyName;
    }

    getCityLeaders(): Array<ICharacter> {
        const citiesAreLoaded = globalThis.CITIES && globalThis.CITIES.citiesCount > 0;
        if (! citiesAreLoaded) {
            this.warn(`cant get city leaders: CityManager not initialized`); return;
        }
        const cityLeaders: Array<ICharacter> = [];
        for (const city of globalThis.CITIES.getAll()) {
            const cityLeader: ICharacter = city.leader as Character;
            if (cityLeader) { cityLeaders.push(cityLeader); }
        }
        return cityLeaders;
    }

    /* NAME LIST */
    getRandomFamilyName(langId: string): string {
        if(! this.hasNameListLoaded) {
            this.error("NAME LIST NOT LOADED");
            return;
        }
        const familyNames = this.nameList[langId][2];
        return RandomElement.of(familyNames);
    }

    /* CHARACTER CREATION */
    addCharacter(characterObj: Character): Character {
        this.characters.push(characterObj);
        return characterObj;
    }

    createCharacterIn(city: City, isLeader?: boolean): Character {
        const charNum = this.characters.length;
        const isFemale = city.num % 2 == 0;
        const cityLanguage = city.language;
        // city is owned by a faction?
        let cityOwnerNum = -1;
        if ( city.owner && city.owner.num >= 0) { cityOwnerNum = city.owner.num; }
        // create new Character object
        let rndChar: Character = new Character(charNum, cityOwnerNum, cityLanguage, null, isFemale);
        // assign home & pos
        rndChar.home = city;
        rndChar.pos = city.pos;
        // add to city characters and make leader when required
        if ( isLeader ) { city.leader = rndChar; }
        city.localCharacters.push(rndChar);
        // add to character list
        this.addCharacter(rndChar);
        return rndChar;
    }

    /* EVENT HANDLING */
    onEvent(ev: IBaseEvent): boolean {
        const success = super.onEvent(ev);
        // cities loaded?
        if (ev.id == EV.ID_CITY_LIST_LOADED) {
            this.log("CITY LIST LOADED");
            this.createCityLeaders();
        }
        // language file loaded?
        if (ev.id == EV.ID_FILE_JSON_LOADED) {
            if (ev.properties.getStr("shortPath") == "default-languages") {
                const languageJson = globalThis.FILES.jsonContent["default-languages"];
                this.onLanguagesLoaded(languageJson);
            }
        }
        return true;
    }

}
