// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// gui
import { ControllerButtons } from "../gui/menus/controller-buttons";
import { EmptyDialog, IDialog } from "../gui/dialogs/common/base-dialog";
import { IGuiContainer, AbstractGuiContainer } from "../gui/common/layouts";
import { DomNode, DomNodes } from "../gui/elements/controls-dom";
import * as GUI_CONST from '../gui/common/constants';
// gui menues
import { TopBarMenu } from '../gui/menus/top-bar';
import { CircularButton } from "../gui/menus/circular-button";
// babylon
import * as GUI from 'babylonjs-gui';
import { AbstractMenu } from '../gui/menus/common/abstract-menu';

const GUI_TREE_DEPTH = 5;

export class GuiManager extends AbstractGuiContainer {
    advTexture: GUI.AdvancedDynamicTexture;
    // menu controls
    menues: Array<AbstractMenu> = [];
    topBar: TopBarMenu;
    controllerBtns: ControllerButtons;

    constructor() {
        super("GuiManager");
        this._createDomRoot();
        this.log(`created`);
    }

    private _createDomRoot() {
        // create global DOM_NODES array
        globalThis.DOM_NODES = new DomNodes();
        // create root node
        const domRoot = new DomNode("root");
        globalThis.DOM_NODES.registerNode(domRoot);
        this.log(`DOM Root created: ${domRoot}`);
    }

    public init(): void {
        this.log("building; doing nothing yet");
    }

    private buildAll(): void {

    }

    getControlById(controlId: string): GUI.Control {
        if (! this.advTexture) {
            this.error(`cant get control "${controlId}"; AdvTexture not found`); return;
        }
        for (const control of this.advTexture.getDescendants()) {
            const hasSameNameOrId: boolean = control.name == controlId;
            if (hasSameNameOrId) { return control; }
        }
        return;
    }

    public showDialog(dialog: IDialog, forceClose?: boolean) {
        this.log(`showing dialog ${dialog.id} ..`);
        this.closeCurrentDialog(forceClose);
        globalThis.DIALOG = dialog;
        globalThis.DIALOG.build();
    }

    public closeCurrentDialog(force?: boolean) {
        if (globalThis.DIALOG) {
            if (globalThis.DIALOG.isBlocking) {
                if (! force ) {
                    this.warn(`cant close current dialog ${globalThis.DIALOG.id}; is blocked`);
                    return;
                }
            }
            globalThis.DIALOG.close();
        }
    }

    /* called from after app assets are loaded */
    public build(): GUI.AdvancedDynamicTexture {
        const scene = globalThis.SCENE;
        this.log(`creating babylon adv texture with scene ${scene}..`);
        this.advTexture = GUI.AdvancedDynamicTexture.CreateFullscreenUI("UI_ROOT", true, scene);
        return this.advTexture;
    }

    showStartingMenues() {
        this.log("showing menues");
        this._createTopBar();
        this._createController();
        this._createCircularButtons();
        this._createBottomMenues();
    }

    /* TOP BAR */
    private _createTopBar(): void {
        this.topBar = new TopBarMenu();
        this.menues.push(this.topBar);
        this.advTexture.addControl(this.topBar);
        this.updateTopBar();
    }

    /* create the gameboy like controller buttons to the bottom right */
    private _createController() {
        this.controllerBtns = new ControllerButtons();
        this.advTexture.addControl(this.controllerBtns);
    }

    private _createCircularButtons() {
        // add next turn button
        const nextTurnBtn = new CircularButton("turn", GUI_CONST.ALIGN_H_RIGHT);
        this.advTexture.addControl(nextTurnBtn);
        const buildBtn = new CircularButton("build", GUI_CONST.ALIGN_H_LEFT);
        this.advTexture.addControl(buildBtn);
    }

    onPlayerMoved() {
        this.updateTopBar();
    }

    updateTopBar(): void {
        this.topBar.updateAll();
    }

    /* BOTTOM ACTIONS */
    private _createBottomMenues() {
        
    }

    playSound(soundName: string): void {

    }

    /* EVENT LISTENING */
    onEvent(ev: IBaseEvent): void {
        super.onEvent(ev);
        if (ev.id == EV.ID_GUI_ELEM_CLICKED) {
            const elemId = ev.properties.getStr("elemId");
            this.onElementClicked(elemId);
        }
    }

    onElementClicked(elemId: string): boolean {
        this.log(`element "${elemId}" clicked!`);
        // trigger onElementClicked in every menu
        for (const menu of this.menues) {
            menu.onElementClicked(elemId);
        }
        // .. & current dialog
        if (globalThis.DIALOG) {
            globalThis.DIALOG.onElementClicked(elemId);
        }
        return false;
    }
}

// == CONSTANTS == */
// MENUES
export const MENU_TOP_ITEM_WIDTH = "150px";