// common
import { Size, WorldPos } from '../world/pos';
import { GameOptions } from '../game-options';
// events
import { IBaseEvent, BaseEventListener } from '../../common/events';
import * as EV from "../game-events";
// world
import { WorldMap } from '../world/map';
import { WorldTile } from '../world/tile/tile';
import { TerrainTypes } from '../world/tile/tile-terrain';
import { LandscapeFeature } from '../world/feature';
import { TargetCursor } from '../world/cursor';
// entities
import { Player } from '../entities/player/player';
import { City } from '../entities/city/city';
import { UnitArmy } from '../entities/unit/army/army';
// helpers
import { CityGenerator, CityLoader } from '../managers/world-loaders';
import { WorldTools } from '../../helpers/world-tools';
// babylon
import * as BABYLON from 'babylonjs';

export interface IMapMaterials {
    grass: BABYLON.StandardMaterial;
    forest: BABYLON.StandardMaterial;
    stone: BABYLON.StandardMaterial;
    mountain: BABYLON.StandardMaterial;
    cave: BABYLON.StandardMaterial;
    mud: BABYLON.StandardMaterial;
    dirt: BABYLON.StandardMaterial;
    water: BABYLON.StandardMaterial;
}

export default class WorldMapScene extends BaseEventListener {
    // basic properties
    private _canvas: HTMLCanvasElement;
    engine: BABYLON.Engine;
    private _light: BABYLON.Light;
    // assets
    private _blueprints = {};
    private tileInstances: Array<BABYLON.InstancedMesh> = [];
    private featureMeshes: Array<BABYLON.AbstractMesh> = [];
    private armyMeshes: Array<BABYLON.AbstractMesh> = [];
    cursorMesh: BABYLON.AbstractMesh;
    // flags
    private assetsLoaded: boolean = false;

    constructor(canvasElement : string) {
        super("WorldScene");
        // Create canvas and engine.
        this._canvas = document.getElementById(canvasElement) as HTMLCanvasElement;
        this.engine = new BABYLON.Engine(this._canvas, true);
    }

    make(): void {
        // make tilemap model
        this.createMapObject();
        // create a basic BJS Scene object.
        globalThis.SCENE = new BABYLON.Scene(this.engine);
        this.delegateInput();
        // create materials & load models
        // this.startAssetLoading();
        globalThis.MAP.createMaterials();
        this.createWorld();
    }

    createWorld() {
        // create camera
        this.placePlayer();
        this.createPlayerCam();
        // create sun
        const sunId = WorldMapScene.NODE_ID_SUN;
        this._light = new BABYLON.HemisphericLight(sunId, new BABYLON.Vector3(0,1,0), globalThis.SCENE);
        // init tiles
        this.makeTileBlueprints(TerrainTypes.LIST);
        this.loadOrGenerateLocations();
        // show map
        // this.showMap(globalThis.MAP);
    }

    private createPlayerCam(): void {
        const player = globalThis.PLAYER;
        if (! player) {
            this.error(`cant create camera; player not set`);
        } else {
            this.log(`camera follows player at ${player.pos}`);
        }
        globalThis.CAMERA = new BABYLON.ArcRotateCamera(
            WorldMapScene.NODE_ID_CAMERA, 
            -Math.PI / 2, Math.PI / 4, 15, 
            BABYLON.Vector3.Zero(), 
            globalThis.SCENE);

        this.renewCameraTarget(player.mesh);
        globalThis.CAMERA.attachControl(this._canvas, true);
    }

    private delegateInput() {
        globalThis.SCENE.onKeyboardObservable.add((kbInfo) => {
            switch (kbInfo.type) {
              case BABYLON.KeyboardEventTypes.KEYDOWN:
                // console.log("KEY DOWN: ", kbInfo.event.key);
                break;
              case BABYLON.KeyboardEventTypes.KEYUP:
                globalThis.INPUT.onKeyUp(kbInfo.event.key);
            }
          });
    }

    private placePlayer(): void {
        this.log(`creating player ..`);
        globalThis.PLAYER = new Player();
    }

    private makeTileBlueprints(terrainTypes: Array<string>): void {
        this.log(`making blueprints ..`);
        for (const terrainId of terrainTypes) {
            this._blueprints[terrainId] = this.makeTileBlueprint(terrainId);
        }
	}

    /* creates Babylon hexagon Mesh for every terrain tile, that will be used
       to make mesh instances of */
    private makeTileBlueprint(terrainType: string): BABYLON.Mesh {
		var diameterBottom = WorldTile.HEXSIZE;
        const tileName = `base-tile-${terrainType}`;
        let mesh: BABYLON.Mesh = BABYLON.Mesh.CreateCylinder(
            tileName,
            TerrainTypes.heightOf(terrainType),
            TerrainTypes.topDiameterOf(terrainType),
            diameterBottom,
            6, // tesselation
            1, // subdivision
            globalThis.SCENE,
            false // updatable
        );
        mesh.rotation.y = Math.PI/2;
		mesh.position.y = TerrainTypes.heightOf(terrainType) /2;
		mesh.convertToFlatShadedMesh();
		mesh.layerMask = 0;
        mesh.material = this.getMaterial(terrainType);
        const viewDist = globalThis.OPTIONS.getNum(GameOptions.MAP_VIEW_DISTANCE);
        mesh.addLODLevel(viewDist, null);
        if (mesh.material) {
            mesh.material.freeze();
        }
        mesh.freezeWorldMatrix();
        return mesh;
    }

    private getMaterial(key: string): BABYLON.StandardMaterial {
        return globalThis.MAP.mats.get(key);
    }

    renewCameraTarget(mesh: BABYLON.AbstractMesh) {
        globalThis.CAMERA.setTarget(mesh);
        // turn to right angle 
        globalThis.CAMERA.radius = CAMERA_RADIUS;
        globalThis.CAMERA.alpha = CAMERA_ALPHA;
        globalThis.CAMERA.beta = CAMERA_BETA;
    }

    private createMapObject(): WorldMap {
        // get map name and default size
        const mapName = globalThis.MAP_NAME;
        const worldSize = new Size(100, 100);
        this.log(`creating map ${mapName}`);
        // generate and return WorldMap model
        globalThis.MAP = new WorldMap(mapName, worldSize);
        console.log(globalThis.MAP);
        return globalThis.MAP;
    }

    private placeNewMap(worldMap: WorldMap): void {
        this.log(`showing map`);
        this.placeMapTiles(worldMap);
        this.placeTargetCursor();
        // loading citied & other locations
        this.placeLocations();
        // place landcape features
        this.placeLandscapeFeatures();
        // place armies & traders
        this.placeOutsideArmies();
        // replace player mesh
        globalThis.PLAYER.updateMesh();
        this.renewCameraTarget(globalThis.PLAYER.mesh);
    }

    /* PLACE MAP TILES */
    private placeMapTiles(worldMap: WorldMap) {
        for (let y = 0; y < worldMap.size.z; y++) {
            for (let x = 0; x < worldMap.size.x; x++) {
                let tile: WorldTile = worldMap.tileAt(x, y);
                this.placeSingleMapTile(tile);
            }
        }
        this.log(`placed ${this.tileInstances.length} MapTiles`);
    }

    private placeSingleMapTile(tile: WorldTile): BABYLON.InstancedMesh {
        if (! tile) { return; }
        // offset every second row
        const offset = (tile.pos.z%2 === 0) ? WorldTile.HEXSIZE/2 : 0;
        // show tile mesh
        const tileInstance = this.createTileInstance(tile, offset);
        this.tileInstances.push(tileInstance);
        return tileInstance;
    }

    private getBlueprint(terrainChar: string): BABYLON.Mesh {
        if (! terrainChar) { terrainChar = TerrainTypes.UNKNOWN; }
        const blueprint: BABYLON.Mesh = this._blueprints[terrainChar];
        if (! blueprint) {
            this.error(`cant create tile with terrain ${terrainChar} in ${this._blueprints}`); return;
        }
        return blueprint;
    }

    private createTileInstance(tile: WorldTile, offset: number): BABYLON.InstancedMesh {
        const blueprint = this.getBlueprint(tile.terrainChar);
        if (! blueprint) {
            this.warn(`cant create tile instance of "${tile.terrainChar}" at ${tile.pos}`); return;
        }
        // create new instance from terrain blueprint
        let meshInstance: BABYLON.InstancedMesh = blueprint.createInstance(tile.meshName);
        meshInstance.position = WorldTools.worldToScenePos(tile.pos);
        // link mesh to tile model
        tile.meshInstance = meshInstance;
        tile.offset = offset;
        return meshInstance;
    }

    get topleft(): WorldPos {
        const map = globalThis.MAP;
        return new WorldPos(-map.size.x * WorldTile.HEXSIZE / 2,
            map.size.z * WorldTile.HEXSIZE / 2);
    }

    getOffsetAt(z: number): number {
        return (z%2 === 0) ? WorldTile.HEXSIZE/2 : 0;
    }

    private loadOrGenerateLocations() {
        this.loadOrGenerateCities();
    }

    private placeLandscapeFeatures(): void {
        const mapObj: WorldMap = globalThis.MAP;
        let placedFeatureCount: number = 0;
        for (const feature of mapObj.features) {
            const successfullyPlaced = this._placeSingleLandscapeFeature( feature );
            if ( successfullyPlaced ) { placedFeatureCount += 1; }
        }
        this.log(`placed ${placedFeatureCount} landscape features on map`);
    }

    private _placeSingleLandscapeFeature(feature: LandscapeFeature): boolean {
        const asset = globalThis.ASSETS.getAsset(feature.id);
        if (! asset) {
            this.warn(`cant place landscape feature "${feature.id}"; asst not found`); return false;
        }
        const meshId = `lds-feature-${this.featureMeshes.length}`;
        const meshInstance = WorldTools.placeAssetInstance(feature.id, meshId, feature.pos, true, true);
        this.featureMeshes.push( meshInstance );
        return meshInstance != null;
    }

    private placeLocations(): void {
        // this.loadOrGenerateCities();
        this.placeCities();
    }

    private loadOrGenerateCities(): void {
        // get loaed or generated City array
        let cities: Array<City> = this.getLoadedOrGeneratedCities();
        // add new cities to CityList holder & inform CityManager
        for (const city of cities) { globalThis.CITIES.addCity(city); }
        globalThis.CITIES.onCitiesLoaded();
    }

    private getLoadedOrGeneratedCities(): Array<City> {
        const generateLocations = globalThis.OPTIONS.getBool(GameOptions.GENERATE_LOCATIONS);
        // generate or load cities
        if (generateLocations) { return this.generateRandomCities(); }
        return this.loadWorldCities();
    }

    private generateRandomCities(): Array<City> {
        this.log(`generating cities ..`);
        let generatedCities: Array<City> = new CityGenerator().generate();
        this.log(`generated ${generatedCities.length} cities`);
        return generatedCities;
    }

    private loadWorldCities(): Array<City> {
        this.log(`loading cities ..`);
        let loadedCities: Array<City> = CityLoader.load();
        this.log(`loaded ${loadedCities.length} cities starting with ${loadedCities[0].name}`);
        return loadedCities;
    }

    private placeCities() {
        const cities = globalThis.CITIES.cities;
        let placedCitiesCount: number = 0;
        this.log(`placing ${cities.length} cities starting with ${cities[0].name} ..`);
        for (const city of cities) {
            this.placeSingleCity(city);
            this.showCityConnections(city);
            placedCitiesCount += 1;
        }
        // fire CitiesPlayedEvent
        globalThis.EVENTS.fireEvent( new EV.CitiesPlacedEvent(placedCitiesCount) );
        // globalThis.CITIES.onCityMeshesPlaced();
    }

    checkAllCitiesPlaces() {

    }

    private placeSingleCity(city: City) {
        const cityAssetName = city.getAssetName();
        city.mesh = globalThis.ASSETS.getAsset(cityAssetName).getMeshInstance(city.meshId, true);
        WorldTools.placeMesh(city.mesh, city.pos, true, true);
    }

    private showCityConnections(city: City) {
        if (! SHOW_CONNECTION_LINES) { return; }
        const sourceVec3 = WorldTools.worldToScenePos(city.pos);
        sourceVec3.y = 0.9;
        let targetVec3Arr: Array<BABYLON.Vector3> = [];
        for (const connectedCities of city.getConnectedTradeLocations()) {
            const targetPos = WorldTools.worldToScenePos(connectedCities.pos);
            targetPos.y = .9;
            targetVec3Arr.push(targetPos);
        }
        let connectionNum: number = 0;
        for (const targetVec3 of targetVec3Arr) {
            const lineOptions = { points: [sourceVec3, targetVec3] };
            // create line
            const connectionLineId = "conn-" + city.key + "-" + connectionNum;
            let line: BABYLON.LinesMesh = BABYLON.MeshBuilder.CreateLines(connectionLineId, lineOptions, globalThis.SCENE);
            line.alpha = 0.3;
            // line.edgesRenderer = new BABYLON.LineEdgesRenderer(line);
            line.edgesWidth = 3;
            line.freezeWorldMatrix();
            connectionNum += 1;
        }

    }

    /* == ARMY PLACEMENT == */
    placeOutsideArmies() {
        const outsideArmies = globalThis.ARMIES.getOutsideArmies();
        for (const army of outsideArmies) {
            this.placeSingleArmy(army);
        }
    }

    placeSingleArmy(army: UnitArmy) {
        const assetName = UnitArmy.DEFAULT_MODEL;
        const meshId = "army-" + army.num;
        const instance = WorldTools.placeAssetInstance(assetName, meshId, army.pos, true, true);
        // add to army meshes holder
        this.armyMeshes.push(instance);
    }

    placeTargetCursor() {
        const cursorDefaultAssetName = TargetCursor.DEFAULT_ASSET_NAME;
        const cursorStartPos: WorldPos = new WorldPos(20, 20);
        this.cursorMesh = WorldTools.placeAssetInstance(cursorDefaultAssetName, TargetCursor.NODE_ID, cursorStartPos, true, true);
    }

    onAllAssetsLoaded() {
        this.placeNewMap(globalThis.MAP);
    }

    doRender(): void {
        // Run the render loop.
        this.engine.runRenderLoop(() => {
            globalThis.SCENE.render();
        });

        // The canvas/window resize event handler.
        window.addEventListener('resize', () => {
            this.log("resized");
            this.engine.resize();
        });
    }

    private static get NODE_ID_CAMERA(): string { return "cam-main"; }
    private static get NODE_ID_SUN(): string { return "sun"; }

    toString(): string {
        let mapName = "no-map";
        let mapSize = "no-size";
        if (globalThis.MAP) {
            mapName = globalThis.MAP.name;
            mapSize = `${globalThis.MAP.tiles[0].length}x${globalThis.MAP.tiles.length}`;
        }
        return `<WorldScene "${mapName}" tiles=${mapSize}> `;
    }
    // events
    onEvent(ev: IBaseEvent) {
        super.onEvent(ev);
    }
}

/* == SCENE CONSTANTS == */
export const SHOW_CITY_MODELS = true;
// camera
export const CAMERA_RADIUS = 25;
export const CAMERA_ALPHA = -1 * Math.PI /2;
export const CAMERA_BETA = 0.84;

export const SHOW_CONNECTION_LINES = false;