// common
import { Properties } from "../../common/properties";
import { DawnbringerPalette } from "../../common/colors";
// events
import { IBaseEvent, BaseEvent, BaseEventListener } from "../../common/events";
import * as EV from "../game-events";
// entities
import { IWorldEntity, WorldEntity, WorldEventEntity } from "../../common/entity";
import { Building } from "../entities/building/building";
import { IEmpire } from "../entities/faction/empire";
import { IFamily } from "../entities/faction/family";
// world
import { IRoad } from "./road";
import { WorldTile } from "./tile/tile";
import { IFaction } from "../entities/faction/faction";
import { WorldPos } from "./pos";
// trade
import { MarketTrade } from "../entities/trade/market-trade";
import { TradeNode } from "../entities/trade/trade-node";
import { TradeConnection } from "../entities/trade/connection/connection";
import { StoredResource } from "../entities/trade/resource/resource-storage";
import { ITradeHistory, ITradeHistoryKeeper, TradeHistory } from "../entities/trade/trade-history";
// babylon 
import * as BABYLON from 'babylonjs';

/* == INTERFACES == */
export interface ILocation extends IWorldEntity {
    // basic properties
    key: string;
    level: number;
    name: string;
    mesh: BABYLON.AbstractMesh;
    properties: Properties;
    money: number;
    // flags
    canTrade: boolean;
    // world links
    roads: Array<IRoad>;
    tile: WorldTile;
    neighborTiles: Array<WorldTile>;
    neighborTerrains: Array<string>;
    terrainId: string;
    // buildings
    buildings: Array<Building>;
    // methods
    equals(otherLocation: ILocation): boolean;
    addNewBuilding(bldType: string);
}

export interface IConnectedLocation extends ILocation {
    // connections
    connections: Array<TradeConnection>;
    // methods
    getConnectionTo(target: IConnectedLocation): TradeConnection;
    addConnectionTo(targetLocation: IConnectedLocation, addMirrored?: boolean): TradeConnection;
    getConnectedLocations(): Array<IConnectedLocation>;
}

export interface IOwnedLocation extends IConnectedLocation {
    owner: IFaction;
    ownerEmpire: IEmpire;
    ownerFamily: IFamily;
    ownerName: string;
    _ownerFactionNum: number;
    _factionKey: string;
    // methods
    getOwner(): IFaction;
}

export interface ITradeLocation extends IOwnedLocation, ITradeHistoryKeeper {
    tradeNode: TradeNode;
    tradeHistory: ITradeHistory;
    // methods
    addTradeConnectionTo(targetLocation: ITradeLocation, addMirrored?: boolean): TradeConnection
    formCaravan(target: ITradeLocation, resId: string);
    canAfford(resId: string, amount: number): boolean;
    getStoredResources(): Array<StoredResource>;
    canProduce(resId: string): boolean;
    produceResource(resId: string, amount: number);
}

/* == ABSTRACT LOCATION MODEL == */
export class AbstractLocation extends WorldEventEntity implements ILocation  {
    // basic properties
    key: string = null;
    level = 1;
    mesh: BABYLON.AbstractMesh;
    name: string;
    properties: Properties = new Properties();
    // flags
    canTrade: boolean = false;
    // world properties
    neighborTiles: WorldTile[] = [];
    roads: IRoad[] = [];
    // buildings
    buildings: Building[] = [];

    constructor(num: number, pos: WorldPos, name: string) {
        super(num, pos);
        this.name = name;
        this.randomizeStartingMoney();
    }

    randomizeStartingMoney() {
        this.properties.addInteger(PROPERTY_MONEY, -1, DawnbringerPalette.GOLDENFIZZ, 1000);
    }

    public equals(otherCity: ILocation) {
        return this.name == otherCity.name;
    }

    set money(value: number) { this.properties.getProperty(PROPERTY_MONEY).value = value; }
    get money(): number { return this.properties.getNum(PROPERTY_MONEY); }

    /* TILE */
    get tile(): WorldTile {
        const tileAt = globalThis.MAP.tileAtPos(this.pos);
        return tileAt;
    }

    get neighborTerrains(): string[] { return this.tile.neighborTerrains; }
    get terrainId(): string { return this.tile.terrainChar; }

    /* BUILDINGS */
    public addNewBuilding(bldType: string, level?: number) {
        // this.log(`adding building ${bldType}`);
        if (! level) { level = 0; }
        const bldInstance: Building = new Building(bldType, "", level);
        this.buildings.push(bldInstance);
        // fire ConstructionFinishedEvent
        globalThis.EVENTS.fireEvent( new EV.ConstructionFinishedEvent( this, bldType, level));
    }
}

export class ConnectedLocation extends AbstractLocation implements IConnectedLocation {
    connections: Array<TradeConnection> = [];

    constructor(num: number, pos: WorldPos, name: string) {
        super(num, pos, name);
    }

    getConnectionTo(target: IConnectedLocation): TradeConnection {
        for (const connection of this.connections) {
            const isToTarget = connection.target.equals(target);
            if (isToTarget) {
                return connection;
            }
        }
        return null;
    }

    addConnectionTo(targetLocation: IConnectedLocation, addMirrored?: boolean): TradeConnection {
        let connection: TradeConnection = new TradeConnection(this, targetLocation);
        this.connections.push(connection);
        // this.tradeNode.linkedMarkets.push(targetLocation.tradeNode);
        if (addMirrored) { targetLocation.addConnectionTo(this, false); }
        return connection;
    }

    getConnectedLocations(): Array<IConnectedLocation> {
        let connected: Array<IConnectedLocation> = [];
        for (const connection of this.connections) {
            if (connection.target.key != this.key) {
                connected.push(connection.target);
            }
        } // end of connections loop
        return connected;
    }


}

/* == OWNED LOCATION MODEL == */
export class OwnedLocation extends ConnectedLocation implements IOwnedLocation {
    owner: IFaction = null;
    _ownerFactionNum: number;
    _familyNum: number;
    _factionKey: string;

    constructor(num: number, pos: WorldPos, name: string, owner?: IFaction) {
        super(num, pos, name);
        if (owner) { this.owner = owner; }
    }

    getOwner(): IFaction {
        return this.owner;
    }

    get ownerEmpire(): IEmpire {
        if (! globalThis.FACTIONS) {
            this.warn(`cant get empire of "${ this.name }"; FactionManager not found`); return;
        }
        return globalThis.FACTIONS.getEmpireByKey(this._factionKey);
    }

    get ownerFamily(): IFamily {
        if (! globalThis.FACTIONS) {
            this.warn(`cant get family of "${ this.name }"; FactionManager not found`); return;
        }
        if (this._familyNum) {
            return globalThis.FACTIONS.getFamilyByNum(this._familyNum);
        }
        return;
    }

    get ownerName(): string {
        if (this.ownerFamily) { return this.ownerFamily.name; }
        if (this.ownerEmpire) { return this.ownerEmpire.name; }
        return "Independent";
    }
    
}

/* == TRADE LOCATION MODELS == */
export class TradeLocation extends OwnedLocation implements ITradeLocation, ITradeHistoryKeeper {
    tradeNode: TradeNode;
    tradeHistory: TradeHistory;
    canTrade: boolean = true;

    constructor(num: number, pos: WorldPos, name: string, owner?: IFaction) {
        super(num, pos, name, owner);
        this.tradeHistory = new TradeHistory(this);
    }

    addTradeConnectionTo(targetLocation: ITradeLocation, addMirrored?: boolean): TradeConnection {
        let connection = super.addConnectionTo(targetLocation, addMirrored);
        this.tradeNode.linked.push(targetLocation.tradeNode);
        return connection;
    }

    tradeIsExport(trade: MarketTrade): boolean { return this.tradeNode.tradeIsExport(trade); }

    /* RESOURCES */
    canAfford(resId: string, amount: number): boolean {
        const singlePrice = this.tradeNode.priceOf(resId);
        const totalPrice = amount * singlePrice;
        if (this.money >= totalPrice) {
            return true;
        }
        return false;
    }

    getStoredResources(): Array<StoredResource> {
        let actuallyStored: Array<StoredResource> = []
        for (const storedRes of this.tradeNode.asStoredResources()) {
            let resAmount = this.tradeNode.amountOf(storedRes.resId);
            if ( resAmount > 0 ) {
                let resPrice: number = this.tradeNode.priceOf(storedRes.resId);
                let stored: StoredResource = new StoredResource(storedRes.resId, resAmount, resPrice);
                actuallyStored.push( stored );
            }
        }
        return actuallyStored;
    }

    canProduce(resId: string): boolean {
        const resObj = globalThis.TRADE.GetRes(resId);
        // can build on city tile terrain?
        if (resObj.canBuildProdBldOn(this.terrainId)) { return true; }
        // can build on neighboring tile terrain?
        for (const terrainChar of this.neighborTerrains) {
            if (resObj.canBuildProdBldOn(terrainChar)) { return true; }
        }
        return false;
    }

    produceResource(resId: string, amount: number) {
        this.tradeNode.add(resId, amount);
    }

    /* CARAVAMS */
    formCaravan(target: ITradeLocation, resId: string, teleport?: boolean) {
        const resName = globalThis.TRADE.getResName(resId);
        this.log(`forming ${resName} caravan to ${target.name}`);
    }

    /* CONNECTIONS */
    getConnectedTradeLocations(): Array<ITradeLocation> {
        let tradeLocations: Array<ITradeLocation> = [];
        for (const connected of this.getConnectedLocations()) {
            if (connected instanceof TradeLocation) {
                tradeLocations.push(connected);
            }
        }
        return tradeLocations;
    }

}

/* == CONSTANTS == */
export const PROPERTY_MONEY = "money";